@@ -152,6 +152,9 @@ class Cog(commands.Cog):
         """Simple command error handler to make error messages utilize [`send_error()`][tidegear.utils.send_error]."""
         if isinstance(error, commands.CommandInvokeError):
             error = error.original
+        if isinstance(error, (commands.MissingRequiredArgument, commands.MissingRequiredAttachment, commands.MissingRequiredFlag)):
+            await self.bot.send_help_for(ctx, help_for=ctx.command)
+            return
         if isinstance(error, commands.BadArgument):
             await send_error(ctx, content=str(error), ephemeral=True)
             return
