@@ -15,7 +15,7 @@ if not find_spec("class_registry"):
 
 from .cog import SentinelCog
 from .db import AbstractPartial, Change, Moderation, PartialChannel, PartialGuild, PartialUser
-from .exceptions import HandlerError, LoggedHandlerError
+from .exceptions import HandlerError, LoggedHandlerError, UpsertError
 from .type import ModerationType, moderation_type_registry
 
 __all__ = [
@@ -30,6 +30,7 @@ __all__ = [
     "moderation_type_registry",
     "HandlerError",
     "LoggedHandlerError",
+    "UpsertError",
 ]
 
 warn("Tidegear Sentinel is still a heavy work-in-progress, and it has been imported!")
