@@ -5,22 +5,42 @@
 """Sentinel database table models."""
 
 from abc import abstractmethod
-from datetime import UTC, datetime
+from datetime import UTC, datetime, timedelta
+from decimal import Decimal
 from enum import StrEnum
-from typing import Any, Literal, Self, overload
+from typing import TYPE_CHECKING, Any, Literal, Self, overload
 
 import orjson
 import rich.repr
 from class_registry.base import RegistryKeyError
 from discord import Guild, Member, Object, Thread, User, abc
-from piccolo.columns.column_types import JSON, Boolean, ForeignKey, Integer, Interval, Serial, Text, Timestamptz, Varchar
+from piccolo.columns.column_types import JSON, Boolean, ForeignKey, Integer, Serial, Text, Timestamptz, Varchar
 from piccolo.columns.defaults.timestamptz import TimestamptzNow
 from piccolo.table import Table as BaseTable
 from redbot.core.bot import Red
 from typing_extensions import override
 
+from tidegear.exceptions import NotFoundError
+from tidegear.sentinel.exceptions import NotReadyError, UpsertError
 from tidegear.sentinel.type import ModerationType, moderation_type_registry
 
+if TYPE_CHECKING:
+    from tidegear.sentinel.cog import SentinelCog
+
+
+def _clean_foreign_keys(obj: dict) -> dict:
+    cleaned = {}
+    for k, v in obj.items():
+        if isinstance(v, dict):
+            nested = _clean_foreign_keys(v)
+            if all(value is None for value in nested.values()):
+                cleaned[k] = None
+            else:
+                cleaned[k] = nested
+        else:
+            cleaned[k] = v
+    return cleaned
+
 
 class Table(BaseTable):
     """Subclass of Piccolo's Table class that allows for easier pretty printing of table rows."""
@@ -39,6 +59,47 @@ class Table(BaseTable):
 
     __rich_repr__.angular = True  # pyright: ignore[reportFunctionMemberAccess]
 
+    @overload
+    def json(self, *, string: Literal[False] = False, indent: bool = ..., options: int = ...) -> bytes: ...
+
+    @overload
+    def json(self, *, string: Literal[True], indent: bool = ..., options: int = ...) -> str: ...
+
+    def json(self, *, string: bool = False, indent: bool = False, options: int = 0) -> bytes | str:
+        """Get a JSON representation of a table object.
+
+        Utilizes a custom serializer that supports the following types, in addition to those supported
+        by the [`orjson`](https://github.com/ijl/orjson?tab=readme-ov-file#serialize) library.
+
+        - [`Table`][tidegear.sentinel.db.tables.Table]
+
+        Args:
+            string: Whether or not to return a string. Avoids a manual decode step.
+            indent: Whether or not to indent the output JSON. Don't use in performance-sensitive environments.
+            options: Additional [options](https://github.com/ijl/orjson?tab=readme-ov-file#option) to pass into `orjson.dumps()`.
+
+        Returns:
+            (bytes): When `string` is `False,` the resulting JSON in bytes. Use [`.decode()`][bytes.decode] to get a usable string.
+            (str): When `string` is `True,` the resulting JSON in a string.
+        """
+        if indent:
+            options |= orjson.OPT_INDENT_2
+
+        dump = orjson.dumps(self, default=self._json_serializer, option=options)
+
+        if string:
+            return dump.decode("utf-8")
+        return dump
+
+    @staticmethod
+    def _json_serializer(obj: object) -> object:
+        if isinstance(obj, Table):
+            return _clean_foreign_keys(obj.to_dict())
+        if isinstance(obj, Decimal):
+            return float(obj)
+        msg = f"Type {type(obj)} is not serializable!"
+        raise TypeError(msg)
+
 
 class AbstractPartial:
     """An abstract class for Partials, detailing methods that should always be implemented within a Partial.
@@ -144,7 +205,7 @@ class PartialGuild(Table, AbstractPartial):
             guild: The guild object to upsert.
 
         Raises:
-            ValueError: If upserting fails for some reason.
+            UpsertError: If upserting fails for some reason.
 
         Returns:
             (PartialGuild): The resulting PartialGuild object.
@@ -158,7 +219,7 @@ class PartialGuild(Table, AbstractPartial):
         if result := await query:
             return result
         msg = "Upsert operation failed!"
-        raise ValueError(msg)
+        raise UpsertError(msg)
 
 
 class PartialUser(Table, AbstractPartial):
@@ -250,7 +311,7 @@ class PartialUser(Table, AbstractPartial):
             user: The User object to upsert.
 
         Raises:
-            ValueError: If upserting fails for some reason.
+            UpsertError: If upserting fails for some reason.
 
         Returns:
             (PartialUser): The resulting PartialUser object.
@@ -266,7 +327,7 @@ class PartialUser(Table, AbstractPartial):
         if result := await query:
             return result
         msg = "Upsert operation failed!"
-        raise ValueError(msg)
+        raise UpsertError(msg)
 
 
 class PartialChannel(Table, AbstractPartial):
@@ -348,7 +409,7 @@ class PartialChannel(Table, AbstractPartial):
                 This will have no effect if you haven't ran this method yet on this PartialChannel instance.
 
         Raises:
-            ValueError: If the guild tied to this channel no longer exists in the database.
+            NotFoundError: If the guild tied to this channel no longer exists in the database.
                 This should be reported as a bug if it occurs, as there's a foreign key constraint that should prevent this on the table itself.
 
         Returns:
@@ -359,7 +420,7 @@ class PartialChannel(Table, AbstractPartial):
 
         if not (guild := await PartialGuild.objects().where(PartialGuild.id == self.guild_id).first()):
             msg = f"No guild exists in the database with id {self.guild_id}!"
-            raise ValueError(msg)
+            raise NotFoundError(msg)
 
         self._guild_obj = guild
         return self._guild_obj
@@ -372,7 +433,7 @@ class PartialChannel(Table, AbstractPartial):
             channel: The channel object to upsert.
 
         Raises:
-            ValueError: If upserting fails for some reason.
+            UpsertError: If upserting fails for some reason.
 
         Returns:
             (PartialChannel): The resulting PartialChannel object.
@@ -388,7 +449,7 @@ class PartialChannel(Table, AbstractPartial):
         if result := await query:
             return result
         msg = "Upsert operation failed!"
-        raise ValueError(msg)
+        raise UpsertError(msg)
 
 
 class Change(Table):
@@ -416,13 +477,17 @@ class Change(Table):
     timestamp: Timestamptz = Timestamptz(default=datetime.now, null=False)
     moderator_id: ForeignKey[PartialUser] = ForeignKey(references=PartialUser, null=False)
     reason: Text = Text(default=None, null=True)
-    duration: Interval = Interval(default=None, null=True)
+    end_timestamp: Timestamptz = Timestamptz(default=None, null=True)
 
     @property
-    def end_timestamp(self) -> datetime | None:
-        """Return the datetime at which the moderation associated with this change will expire, if this change is the latest change."""
-        if self.timestamp and self.duration:
-            return self.timestamp + self.duration
+    def duration(self) -> timedelta | None:
+        """Retrieve the timedelta between the change's timestamp and end timestamp.
+
+        Returns:
+            The difference (timedelta) between the end timestamp and the timestamp.
+        """
+        if self.timestamp and self.end_timestamp:
+            return self.end_timestamp - self.timestamp
         return None
 
     async def moderation(self, /, *, cache: bool = True) -> "Moderation":
@@ -433,7 +498,7 @@ class Change(Table):
                 This will have no effect if you haven't ran this method yet on this Change instance.
 
         Raises:
-            ValueError: If the [`Moderation`][tidegear.sentinel.Moderation] tied to this change no longer exists in the database.
+            NotFoundError: If the [`Moderation`][tidegear.sentinel.Moderation] tied to this change no longer exists in the database.
                 This should be reported as a bug if it occurs, as there's a foreign key constraint that should prevent this on the table itself.
 
         Returns:
@@ -444,7 +509,7 @@ class Change(Table):
 
         if not (mod := await Moderation.objects().where(Moderation.id == self.moderation_id).first()):
             msg = f"Moderation with id {self.moderation_id} does not exist in the database!"
-            raise ValueError(msg)
+            raise NotFoundError(msg)
 
         self._moderation_obj = mod
         return mod
@@ -457,7 +522,7 @@ class Change(Table):
                 This will have no effect if you haven't ran this method yet on this Change instance.
 
         Raises:
-            ValueError: If the [`PartialUser`][tidegear.sentinel.PartialUser] tied to this change no longer exists in the database.
+            NotFoundError: If the [`PartialUser`][tidegear.sentinel.PartialUser] tied to this change no longer exists in the database.
                 This should be reported as a bug if it occurs, as there's a foreign key constraint that should prevent this on the table itself.
 
         Returns:
@@ -468,7 +533,7 @@ class Change(Table):
 
         if not (user := await PartialUser.objects().where(PartialUser.id == self.moderator_id).first()):
             msg = f"PartialUser with id {self.moderator_id} does not exist in the database!"
-            raise ValueError(msg)
+            raise NotFoundError(msg)
 
         self._moderator_obj = user
         return user
@@ -482,7 +547,7 @@ class Moderation(Table):
         guild_id: The internal ID of the guild this moderation case originates from within the database.
             **This is not a Discord Guild ID!**
             Use [`.guild()`][tidegear.sentinel.Moderation.guild] to get an actual [`PartialGuild`][tidegear.sentinel.PartialGuild] object.
-        timestamp: The datetime at which this moderation occurred.
+        timestamp: A timezone-aware datetime at which this moderation occurred.
         type_key: The moderation type this moderation uses.
             Use [`.type`][tidegear.sentinel.Moderation.type] to get an actual [`ModerationType`][tidegear.sentinel.ModerationType] object.
         target_user_id: The internal ID of the user this moderation case targets within the database.
@@ -496,7 +561,7 @@ class Moderation(Table):
         moderator_id: The internal ID of the user this moderation case was created by within the database.
             **This is not a Discord User ID!**
             Use [`.moderator()`][tidegear.sentinel.Moderation.moderator] to get an actual [`PartialUser`][tidegear.sentinel.PartialUser] object.
-        duration: A timedelta for how long this moderation should last.
+        end_timestamp: A timezone-aware datetime at which this moderation should expire.
         expired: A boolean for if this moderation has expired yet.
         reason: The reason associated with this moderation.
         resolved: A boolean for if this moderation has been resolved.
@@ -515,7 +580,7 @@ class Moderation(Table):
     target_user_id: ForeignKey[PartialUser] = ForeignKey(references=PartialUser, null=True, index=True)
     target_channel_id: ForeignKey[PartialChannel] = ForeignKey(references=PartialChannel, null=True, index=True)
     moderator_id: ForeignKey[PartialUser] = ForeignKey(references=PartialUser, null=False, index=True)
-    duration: Interval = Interval(default=None, null=True)
+    end_timestamp: Timestamptz = Timestamptz(default=None, null=True, index=True)
     expired: Boolean = Boolean(default=False, null=False)
     reason: Text = Text(default=None, null=True)
     resolved: Boolean = Boolean(default=False, null=False)
@@ -524,18 +589,18 @@ class Moderation(Table):
     metadata: JSON = JSON(default="{}", null=False)
 
     @property
-    def end_timestamp(self) -> datetime | None:
-        """Retrieve the datetime at which the moderation should expire, if it has a duration.
+    def duration(self) -> timedelta | None:
+        """Retrieve the timedelta between the moderation's timestamp and end timestamp.
 
         Warning:
             This property does not check if the moderation's type supports expiry.
             Instead, use [`.type.can_expire`][tidegear.sentinel.ModerationType.can_expire] for that.
 
         Returns:
-            The datetime at which the moderation should expire, or `None` if the moderation does not have a duration set.
+            The difference (timedelta) between the end timestamp and the timestamp.
         """
-        if self.duration:
-            return self.timestamp + self.duration
+        if self.end_timestamp:
+            return self.end_timestamp - self.timestamp
         return None
 
     @property
@@ -573,7 +638,7 @@ class Moderation(Table):
         if cache and "_changes" in self.__dict__:
             return self._changes
 
-        changes = await Change.objects().where(Change.moderation_id == self.id)
+        changes = await Change.objects(Change.all_related()).where(Change.moderation_id == self.id)
 
         self._changes = changes
         return changes
@@ -586,7 +651,7 @@ class Moderation(Table):
                 This will have no effect if you haven't ran this method yet on this Moderation instance.
 
         Raises:
-            ValueError: If the [`PartialGuild`][tidegear.sentinel.PartialGuild] tied to this moderation no longer exists in the database.
+            NotFoundError: If the [`PartialGuild`][tidegear.sentinel.PartialGuild] tied to this moderation no longer exists in the database.
                 This should be reported as a bug if it occurs, as there's a foreign key constraint that should prevent this on the table itself.
 
         Returns:
@@ -597,7 +662,7 @@ class Moderation(Table):
 
         if not (guild := await PartialGuild.objects().where(PartialGuild.id == self.guild_id).first()):
             msg = f"Could not find a PartialGuild in the database with id {self.guild_id}"
-            raise ValueError(msg)
+            raise NotFoundError(msg)
 
         self._guild_obj = guild
         return guild
@@ -610,7 +675,7 @@ class Moderation(Table):
                 This will have no effect if you haven't ran this method yet on this Moderation instance.
 
         Raises:
-            ValueError: If the [`PartialUser`][tidegear.sentinel.PartialUser] tied to this moderation no longer exists in the database.
+            NotFoundError: If the [`PartialUser`][tidegear.sentinel.PartialUser] tied to this moderation no longer exists in the database.
                 This should be reported as a bug if it occurs, as there's a foreign key constraint that should prevent this on the table itself.
 
         Returns:
@@ -621,7 +686,7 @@ class Moderation(Table):
 
         if not (user := await PartialUser.objects().where(PartialUser.id == self.moderator_id).first()):
             msg = f"Could not find a PartialUser in the database with id {self.moderator_id}"
-            raise ValueError(msg)
+            raise NotFoundError(msg)
 
         self._moderator_obj = user
         return user
@@ -634,9 +699,10 @@ class Moderation(Table):
                 This will have no effect if you haven't ran this method yet on this Moderation instance.
 
         Raises:
-            ValueError: If the [`PartialUser`][tidegear.sentinel.PartialUser] or [`PartialChannel`][tidegear.sentinel.PartialChannel]
+            NotFoundError: If the [`PartialUser`][tidegear.sentinel.PartialUser] or [`PartialChannel`][tidegear.sentinel.PartialChannel]
                 tied to this moderation no longer exists in the database.
                 This should be reported as a bug if it occurs, as there's a foreign key constraint that should prevent this on the table itself.
+            ValueError: If neither `target_user` or `target_channel` are set on the moderation object.
 
         Returns:
             The PartialUser or PartialChannel that was targeted by this moderation.
@@ -647,12 +713,12 @@ class Moderation(Table):
         if self.target_user_id is not None:
             if not (result := await PartialUser.objects().where(PartialUser.id == self.target_user_id).first()):
                 msg = f"Could not find a PartialUser in the database with id {self.target_user_id}"
-                raise ValueError(msg)
+                raise NotFoundError(msg)
 
         elif self.target_channel_id is not None:
             if not (result := await PartialChannel.objects().where(PartialChannel.id == self.target_channel_id).first()):
                 msg = f"Could not find a PartialChannel in the database with id {self.target_channel_id}"
-                raise ValueError(msg)
+                raise NotFoundError(msg)
 
         else:
             msg = "Neither target_user nor target_channel are set!"
@@ -669,7 +735,7 @@ class Moderation(Table):
                 This will have no effect if you haven't ran this method yet on this Moderation instance.
 
         Raises:
-            ValueError: If the [`PartialUser`][tidegear.sentinel.PartialUser] tied to this moderation no longer exists in the database.
+            NotFoundError: If the [`PartialUser`][tidegear.sentinel.PartialUser] tied to this moderation no longer exists in the database.
                 This should be reported as a bug if it occurs, as there's a foreign key constraint that should prevent this on the table itself.
 
         Returns:
@@ -680,17 +746,18 @@ class Moderation(Table):
 
         if not (user := await PartialUser.objects().where(PartialUser.id == self.resolver_id).first()):
             msg = f"Could not find a PartialUser in the database with id {self.resolver_id}"
-            raise ValueError(msg)
+            raise NotFoundError(msg)
 
         self._resolver_obj = user
         return user
 
-    async def expire(self) -> Self:
+    async def expire(self, cog: "SentinelCog") -> Self:
         """Mark a moderation as expired. This will run the moderation type's expiration handler.
 
         Raises:
             ValueError: If the moderation is already expired.
             NotImplementedError: If the moderation type does not support expiry.
+            NotReadyError: If the moderation isn't yet ready to expire.
 
         Returns:
             The expired moderation.
@@ -699,14 +766,35 @@ class Moderation(Table):
             msg = f"Moderation {self.id:,} is already expired!"
             raise ValueError(msg)
 
-        if self.type.can_expire:
-            if self.end_timestamp and datetime.now(tz=UTC) >= self.end_timestamp:
-                await self.type.expiry_handler(moderation=self)
+        if self.type.can_expire and self.end_timestamp:
+            if datetime.now(tz=UTC) >= self.end_timestamp:
+                await self.type.expiry_handler(cog, moderation=self)
                 await self.update_self({Moderation.expired: True})
                 return self
-        msg = f"Moderation of type {self.type.key} is not expirable!"
+            msg = f"Moderation {self.id:,} is not ready to expire yet!"
+            raise NotReadyError(msg)
+        msg = f"Moderation of type {self.type.key} is not expirable or does not have a duration!"
         raise NotImplementedError(msg)
 
+    @classmethod
+    async def from_id(cls, moderation_id: int) -> Self:
+        """Retrieve a moderation case by ID.
+
+        Args:
+            moderation_id: The ID of the moderation case to look up.
+
+        Raises:
+            NotFoundError: If the database does not contain a moderation case matching the given ID.
+
+        Returns:
+            The moderation that matches the given ID.
+        """
+        moderation = await cls.objects(cls.all_related()).where(cls.id == moderation_id).first()
+        if not moderation:
+            msg = f"Could not find moderation within the database with an ID of {moderation_id}."
+            raise NotFoundError(msg)
+        return moderation
+
     @classmethod
     async def delete_for_guild(cls, guild: Guild | PartialGuild) -> list[Self]:
         """Delete all Moderation cases for a specific guild.
