@@ -5,16 +5,20 @@
 """Defines the base class for Sentinel moderation types."""
 
 from types import CoroutineType
-from typing import TYPE_CHECKING, Any, Callable
+from typing import TYPE_CHECKING, Any, Callable, Mapping
 
+from class_registry.base import RegistryKeyError
 from class_registry.registry import ClassRegistry
-from discord import Interaction, Member, Permissions, Thread, User, abc
+from discord import Member, Permissions, Thread, User, abc
+from discord.utils import MISSING
 from redbot.core import commands
 from typing_extensions import override
 
+from tidegear import chat_formatting as cf
 from tidegear.utils import class_overrides_attribute
 
 if TYPE_CHECKING:
+    from tidegear.sentinel.cog import SentinelCog
     from tidegear.sentinel.db.tables import Moderation
 
 
@@ -25,11 +29,10 @@ class ModerationType:
         ```python
         from discord import Member, Permissions, User
         from redbot.core import commands
-        from redbot.core.utils import chat_formatting as cf
         from typing_extensions import override
 
-        from tidegear.sentinel.db import Moderation, PartialGuild, PartialUser
-        from tidegear.sentinel.type import ModerationType
+        from tidegear import chat_formatting as cf
+        from tidegear.sentinel import SentinelCog, Moderation, PartialGuild, PartialUser, ModerationType
 
 
         class Warn(ModerationType):
@@ -69,7 +72,7 @@ class ModerationType:
 
             @override
             @classmethod
-            async def resolve_handler(cls, moderation: Moderation, reason: str) -> None:
+            async def resolve_handler(cls, cog: SentinelCog, moderation: Moderation) -> None:
                 return
         ```
 
@@ -83,11 +86,14 @@ class ModerationType:
         removes_from_guild: Whether this type's handler removes the target from the guild,
             or if the moderation is expected to occur whenever the user is not in the guild.
             This does not actually remove the target from the guild; the handler method is responsible for that.
+            **Moderation types that remove users from guilds are responsible for contacting users using the
+            [`SentinelCog.contact_target`][tidegear.sentinel.SentinelCog.contact_target] method *before* removing them from the guild.**
             Defaults to `False`.
         permissions: The Discord permissions required for this type's moderation handler to function.
             Defaults to [`Permissions.none`][discord.Permissions.none].
-        history_metadata: A set of metadata keys to make visible in the output of [`SentinelCog.history`][tidegear.sentinel.SentinelCog.history].
-            Defaults to [`set`][].
+        history_metadata: A mapping of metadata keys to make visible in the output of [`SentinelCog.history`][tidegear.sentinel.SentinelCog.history].
+            Values will be passed through the given function before being output in history,
+            to allow for storing raw values and then formatting them later. Defaults to [`dict`][].
     """
 
     key: str = "type"
@@ -96,7 +102,7 @@ class ModerationType:
     embed_desc: str = "been "
     removes_from_guild: bool = False
     permissions: Permissions = Permissions.none()
-    history_metadata: set[str] = set()
+    history_metadata: Mapping[str, Callable[[str], str]] = {}
 
     @property
     def can_edit_duration(self) -> bool:
@@ -105,7 +111,7 @@ class ModerationType:
         Returns:
             If this type supports editing the duration of moderations.
         """
-        return class_overrides_attribute(child=type(self), parent=ModerationType, attribute="edit_duration_handler")
+        return class_overrides_attribute(child=type(self), parent=ModerationType, attribute="duration_edit_handler")
 
     @property
     def can_expire(self) -> bool:
@@ -202,15 +208,15 @@ class ModerationType:
                 return self.member_target_handler
             if not self.can_target_users:
                 if isinstance(target, User) and self.can_target_members:
-                    msg = f"Moderation type {type(self)} only supports targeting members of the current guild!"
+                    msg = f"Moderation type {self.__class__.__name__} only supports targeting members of the current guild!"
                 else:
-                    msg = f"Moderation type {type(self)} does not support targeting users!"
+                    msg = f"Moderation type {self.__class__.__name__} does not support targeting users!"
                 raise TypeError(msg)
             return self.user_target_handler
 
         if isinstance(target, (abc.GuildChannel, Thread)):
             if not self.can_target_channels:
-                msg = f"Moderation type {type(self)} does not support targeting channels!"
+                msg = f"Moderation type {self.__class__.__name__} does not support targeting channels!"
                 raise TypeError(msg)
             return self.channel_target_handler
 
@@ -218,15 +224,51 @@ class ModerationType:
         raise TypeError(msg)
 
     @classmethod
-    async def member_target_handler(cls, ctx: commands.Context, target: Member, **kwargs: Any) -> "Moderation":
+    async def convert(cls, ctx: commands.Context, argument: str) -> "type[ModerationType]":
+        """A discord.py [`commands.Converter`][discord.ext.commands.Converter] for fetching moderation types.
+
+        Example:
+            ```python
+            from redbot.core import commands
+            from tidegear.sentinel import ModerationType
+
+
+            @commands.command()
+            async def example(self, ctx: commands.Context, moderation_type: ModerationType) -> None:
+                await ctx.send(content=ModerationType.key)
+            ```
+
+        Args:
+            ctx: The context of the command the converter is being used by.
+            argument: The key of the moderation type to search for.
+
+        Raises:
+            commands.BadArgument: If no moderation type exists in the registry with the given key.
+
+        Returns:
+            The fetched moderation type.
+        """
+        try:
+            return moderation_type_registry.get_class(argument)
+        except RegistryKeyError:
+            msg = f"Couldn't find a moderation type with the key {cf.inline(str(argument))}!"
+            raise commands.BadArgument(msg)
+
+    @classmethod
+    async def member_target_handler(
+        cls, *, cog: "SentinelCog", ctx: commands.Context, target: Member, silent: bool = MISSING, **kwargs: Any
+    ) -> "Moderation":
         """This method should be overridden by any child classes that can target members but **not** users,
         and should retain the same starting keyword arguments.
         If your child class can target people outside of the current guild,
         consider using [`.user_target_handler`][tidegear.sentinel.ModerationType.user_target_handler] instead.
 
         Args:
+            cog: A cog instance of a Sentinel cog.
             ctx: The context of the command.
             target: The target of the moderation.
+            silent: Whether or not to direct message the user.
+                This will be ignored if the type has [`removes_from_guild`][tidegear.sentinel.ModerationType] set to `False`.
             **kwargs (dict[str, Any]): Any additional keyword arguments;
                 will be passed in by the [`SentinelCog.moderate`][tidegear.sentinel.SentinelCog.moderate] function.
 
@@ -236,12 +278,17 @@ class ModerationType:
         raise NotImplementedError
 
     @classmethod
-    async def user_target_handler(cls, ctx: commands.Context, target: Member | User, **kwargs: Any) -> "Moderation":
+    async def user_target_handler(
+        cls, *, cog: "SentinelCog", ctx: commands.Context, target: Member | User, silent: bool = MISSING, **kwargs: Any
+    ) -> "Moderation":
         """This method should be overridden by any child classes that can target users, but should retain the same starting keyword arguments.
 
         Args:
+            cog: A cog instance of a Sentinel cog.
             ctx: The context of the command.
             target: The target of the moderation.
+            silent: Whether or not to direct message the user.
+                This will be ignored if the type has [`removes_from_guild`][tidegear.sentinel.ModerationType] set to `False`.
             **kwargs (dict[str, Any]): Any additional keyword arguments;
                 will be passed in by the [`SentinelCog.moderate`][tidegear.sentinel.SentinelCog.moderate] function.
 
@@ -251,11 +298,14 @@ class ModerationType:
         raise NotImplementedError
 
     @classmethod
-    async def channel_target_handler(cls, ctx: commands.Context, target: abc.GuildChannel | Thread, **kwargs: Any) -> "Moderation":
+    async def channel_target_handler(
+        cls, *, cog: "SentinelCog", ctx: commands.Context, target: abc.GuildChannel | Thread, **kwargs: Any
+    ) -> "Moderation":
         """This method should be overridden by any child classes that can target channels or threads,
             but should retain the same starting keyword arguments.
 
         Args:
+            cog: A cog instance of a Sentinel cog.
             ctx: The context of the command.
             target: The target of the moderation.
             **kwargs (dict[str, Any]): Any additional keyword arguments;
@@ -267,33 +317,35 @@ class ModerationType:
         raise NotImplementedError
 
     @classmethod
-    async def resolve_handler(cls, moderation: "Moderation", reason: str) -> None:
+    async def resolve_handler(cls, cog: "SentinelCog", moderation: "Moderation") -> None:
         """This method should be overridden by any resolvable child classes, but should retain the same keyword arguments.
             If your moderation type should not be resolvable, do not override this.
+            This handler should be called after the moderation is marked as resolved within the database.
 
         Args:
-            moderation: The moderation to resolve.
-            reason: The reason for resolving the moderation.
+            cog: A cog instance of a Sentinel cog.
+            moderation: The moderation being resolved.
         """
         raise NotImplementedError
 
     @classmethod
-    async def expiry_handler(cls, moderation: "Moderation") -> None:
+    async def expiry_handler(cls, cog: "SentinelCog", moderation: "Moderation") -> None:
         """This method should be overridden by any expirable child classes, but should retain the same keyword arguments.
             If your moderation type should not expire, do not override this.
 
         Args:
+            cog: A cog instance of a Sentinel cog.
             moderation: The moderation that is expiring.
         """
         raise NotImplementedError
 
     @classmethod
-    async def duration_edit_handler(cls, interaction: Interaction, old_moderation: "Moderation", new_moderation: "Moderation") -> None:
+    async def duration_edit_handler(cls, ctx: commands.Context, old_moderation: "Moderation", new_moderation: "Moderation") -> None:
         """This method should be overridden by any child classes with editable durations, but should retain the same keyword arguments.
             If your moderation type's duration should not be editable, do not override this.
 
         Args:
-            interaction: The interaction that triggered the duration edit.
+            ctx: The context that triggered the duration edit.
             old_moderation: The old moderation, from before the `/edit` command was invoked.
             new_moderation: The current state of the moderation.
         """
